SystemOrganization addCategory: #'XTream-Core'!

Notification subclass: #InappropriateXtreamMessage
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'XTream-Core'!

Object subclass: #Xtream
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'XTream-Core'!

!Xtream commentStamp: 'nice 12/2/2009 03:55' prior: 0!
A Xtream is a replacement for Xtream class.
It is an abstract class with many messages for reading/writing/positionning streams.
Not all stream will be readable/writeable/positionable.
It is then subclassResponsibility to implement some of the basic messages required:

All:
	contentsSpecies
	contents
Positionable:
	position
	position:
	skip:
Readable:
	next
	next:into:startingAt:
Writeable:
	nextPut:
	next:putAll:startingAt:!

Xtream subclass: #ReadXtream
	instanceVariableNames: 'source endOfStreamAction'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'XTream-Core'!

!ReadXtream commentStamp: 'nice 12/2/2009 03:57' prior: 0!
A ReadXtream is an abstract class for reading a sequence of objects

Instance Variables
	endOfStreamAction:		<Valuable> any class than canUnderstand: #value
	source:		<Object>

endOfStreamAction
	- the endOfStreamAction will be evaluated when an attempt to read past end of the stream will occur.
	
source
	- the source of objects we are streaming onto

!

ReadXtream subclass: #BlockReadXtream
	instanceVariableNames: 'contentsSpecies endOfStreamValue'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'XTream-Core'!

!BlockReadXtream commentStamp: 'nice 12/3/2009 09:46' prior: 0!
A BlockReadXtream reads its input from a Block: it has the Block as its source

Instance Variables
	contentsSpecies	<Class> the collection class used to store our contents
	endOfStreamValue	<Object> an object returned by the source block to indicate the end of stream
!

!BlockReadXtream methodsFor: 'accessing' stamp: 'nice 12/3/2009 14:45'!
contents
	"Do not attempt to rewind, just get the contents from current position"
	^self upToEnd! !

!BlockReadXtream methodsFor: 'accessing' stamp: 'nice 12/3/2009 09:32'!
contentsSpecies
	^contentsSpecies! !

!BlockReadXtream methodsFor: 'accessing' stamp: 'nice 12/3/2009 09:33'!
contentsSpecies: aClass
	"Set the class used to gather this stream contents"
	contentsSpecies := aClass! !

!BlockReadXtream methodsFor: 'accessing' stamp: 'nice 12/3/2009 09:39'!
endOfStreamValue: anObject
	"Set the object returned by the block indicating that an endOfStream occured"
	endOfStreamValue := anObject! !

!BlockReadXtream methodsFor: 'initialize-release' stamp: 'nice 12/3/2009 09:32'!
initialize
	contentsSpecies := Array! !

!BlockReadXtream methodsFor: 'reading' stamp: 'nice 12/3/2009 14:47'!
next
	| element |
	element := source value.
	element == endOfStreamValue ifTrue: [^endOfStreamAction value].
	^element! !

!BlockReadXtream methodsFor: 'reading' stamp: 'nice 12/3/2009 09:44'!
readInto: aCollection startingAt: startIndex count: count
	^self restoreEndOfStreamActionAfter: [
		| index nRead |
		index := startIndex.
		nRead := 0.
		endOfStreamAction := [^nRead].
		[nRead = count] whileFalse: [
			aCollection at: index put: self next.
			index := index + 1.
			nRead := 1 + nRead].
		nRead]! !

BlockReadXtream subclass: #BlockReadXtreamWrapper
	instanceVariableNames: 'origin'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'XTream-Core'!

!BlockReadXtreamWrapper commentStamp: 'nice 12/4/2009 03:00' prior: 0!
A BlockReadXtreamWrapper  reads its input from a Block: it has the Block as its source
More over, it retains a pointer to its origin so as to perform some query

Instance Variables
	origin		<ReadXtream>!

!BlockReadXtreamWrapper methodsFor: 'file API' stamp: 'nice 12/4/2009 03:01'!
close
	^origin close! !

!BlockReadXtreamWrapper methodsFor: 'file API' stamp: 'nice 12/4/2009 03:01'!
closed
	^origin closed! !

!BlockReadXtreamWrapper methodsFor: 'file API' stamp: 'nice 12/4/2009 03:01'!
name
	^origin name! !

!BlockReadXtreamWrapper methodsFor: 'accessing' stamp: 'nice 12/4/2009 03:04'!
origin: aReadXtream
	^origin := aReadXtream! !

ReadXtream subclass: #BufferedReadXtream
	instanceVariableNames: 'buffer position readLimit'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'XTream-Core'!

!BufferedReadXtream commentStamp: 'nice 12/2/2009 04:18' prior: 0!
A BufferedReadXtream is a readOnly stream that buffers reads by chunk of fixed size.
The goal of the buffer is to be more efficient than a StandardFileStream reading/writing characters one by one.

instance variables:
- buffer <Collection> is used as the read buffer
- readLimit <Integer> is used as the read limit into the read buffer
  readLimit always reflect the true file position of underlying OS stream with this invariant:
    (trueFilePosition rem: bufferSize) = readLimit
  Thus the file position corresponding to beginning of buffer is always:
   (trueFilePosition - readLimit)
- position <Integer> is used as the index before next character into the read/write buffer
  It is thus the position relative to beginning of buffer.
  the absolute position versus beginning of file is thus:  trueFilePosition - readLimit + position

BEWARE: try to preserve the invariants exposed above.
!

BufferedReadXtream subclass: #BufferedReadWriteXtream
	instanceVariableNames: 'writeLimit bufferLimit'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'XTream-Core'!

!BufferedReadWriteXtream commentStamp: 'nice 12/2/2009 04:16' prior: 0!
A BufferedReadWriteXtream is a read/write stream that buffers reads and writes by chunk.
The goal of the buffer is to be more efficient than a StandardFileStream reading/writing characters one by one.

instance variables:
- buffer <Collection> is used as the read buffer
- readLimit <Integer> is used as the read limit into the read buffer
  readLimit always reflect the true file position of underlying OS stream with this invariant:
    (trueFilePosition rem: bufferSize) = readLimit
  Thus the file position corresponding to beginning of buffer is always:
   (trueFilePosition - readLimit)
- writeLimit <Integer> is used as the further character written so far in the buffer
  writeLimit = 0 means the buffer has not been written at all (it is clean).
- position <Integer> is used as the index before next character into the read/write buffer
  It is thus the position relative to beginning of buffer.
  the absolute position versus beginning of file is thus:  trueFilePosition - readLimit + position
- bufferLimit <Integer> is (readLimit max: writeLimit) it is an optimization to avoid to repeat this operation

BEWARE: try to preserve the invariants exposed above.

TIMING TESTS:

| strm timing lineCount |
timing := OrderedCollection new: 4.
strm := StandardFileXtream oldFileNamed: (SourceFiles at: 1) fullName.
lineCount := 0.
timing add: ([[[strm atEnd] whileFalse: [lineCount := lineCount + 1. strm nextLine]] timeToRun] ensure: [strm close]).

strm := ReadWriteBuffer oldFileNamed: (SourceFiles at: 1) fullName.
timing add: ([[[strm atEnd] whileFalse: [lineCount := lineCount - 1. strm nextLine]] timeToRun] ensure: [strm close]).
self assert: lineCount = 0.

strm := StandardFileXtream oldFileNamed: (SourceFiles at: 1) fullName.
timing add: ([[[strm next == nil] whileFalse] timeToRun] ensure: [strm close]).

strm := ReadWriteBuffer oldFileNamed: (SourceFiles at: 1) fullName.
timing add: ([[[strm next == nil] whileFalse] timeToRun] ensure: [strm close]).

^timing

MY RESULTS (trunk image 2009/11/18)

-> an OrderedCollection(8762 1806 42296 5941)
8762.0/ 1806 -> 4.8516057585825
42296.0 /5941 -> 7.11934017842114

THIS GIVES A RAW ESTIMATION OF POTENTIAL SPEED UP: 5 to 7.!

!BufferedReadWriteXtream methodsFor: 'private' stamp: 'nice 11/28/2009 03:56'!
availableDataInBuffer
	"Answer the amount of data in buffer than can be read starting from current position."
	
	^bufferLimit - position! !

!BufferedReadWriteXtream methodsFor: 'deprecated' stamp: 'nice 12/4/2009 00:53'!
basicNextPut: aChar
	^self nextPut: aChar! !

!BufferedReadWriteXtream methodsFor: 'private' stamp: 'nice 11/28/2009 15:55'!
bufferLimit
	^bufferLimit! !

!BufferedReadWriteXtream methodsFor: 'private' stamp: 'nice 11/28/2009 03:59'!
checkAvailableDataInBuffer
	"Answer true if data is available in buffer.
	In case of end of buffer, force next buffer read."

	bufferLimit > position ifTrue: [^true].
	^self readNextBuffer! !

!BufferedReadWriteXtream methodsFor: 'private' stamp: 'nice 11/28/2009 03:59'!
clearBuffer
	"Set the internal variables to values signifying there is no data currently in buffer."

	position := bufferLimit := readLimit := writeLimit := 0! !

!BufferedReadWriteXtream methodsFor: 'file API' stamp: 'nice 12/4/2009 02:51'!
close
	self writeBufferToFile.
	^source close! !

!BufferedReadWriteXtream methodsFor: 'file API' stamp: 'nice 12/4/2009 02:33'!
closed
	^source closed! !

!BufferedReadWriteXtream methodsFor: 'writing' stamp: 'nice 12/4/2009 01:44'!
encodeWith: aTextConverter
	self ascii.
	^BlockWriteXtreamWrapper new
		destination: [:e | aTextConverter nextPut: e toStream: self. e];
		target: self;
		yourself! !

!BufferedReadWriteXtream methodsFor: 'file API' stamp: 'nice 12/4/2009 02:51'!
flush
	self writeBufferToFile.
	^source flush! !

!BufferedReadWriteXtream methodsFor: 'testing' stamp: 'nice 12/1/2009 22:00'!
isWriteable
	^true! !

!BufferedReadWriteXtream methodsFor: 'writing' stamp: 'nice 11/28/2009 02:19'!
next: n
	"Return a string with the next n characters of the filestream in it."

	| result |
	"check for fast case: enough data in the buffer"
	self availableDataInBuffer >= n
		ifTrue: [result := buffer copyFrom: position + 1 to position + n.
			position := position + n.
			^result].
	"slow case: create then fill a buffer"
	buffer ifNil: [	self initializeBuffer.].
	^self next: n into: (buffer species new: n) startingAt: 1! !

!BufferedReadWriteXtream methodsFor: 'writing' stamp: 'nice 12/3/2009 23:06'!
next: anInteger putAll: aCollection startingAt: startIndex
	| endPosition bufferSize remainingBytesToWrite |
	buffer ifNil: [	self initializeBuffer.].

	bufferSize := buffer size.
	position := position + 1.
	position > readLimit ifTrue: [
		position > bufferSize ifTrue: [
			self readNextBuffer.
			position := position + 1]].

	endPosition := position + anInteger - 1.
	buffer replaceFrom: position to: (position := endPosition min: bufferSize) with: aCollection startingAt: startIndex.
	writeLimit := writeLimit max: position.
	bufferLimit := writeLimit max: readLimit.
	(remainingBytesToWrite := endPosition - position) > 0 ifTrue: [
		self writeBufferToFile.
		self clearBuffer.
		remainingBytesToWrite > bufferSize ifTrue: [
			"since remainingBytesToWrite is long, bypass the buffer and write directly to file"
			"source writeFrom: aCollection startingAt: startIndex + anInteger - remainingBytesToWrite count: (remainingBytesToWrite quo: bufferSize) * bufferSize."
			source next: (remainingBytesToWrite quo: bufferSize) * bufferSize putAll: aCollection startingAt: startIndex + anInteger - remainingBytesToWrite.
			remainingBytesToWrite := remainingBytesToWrite rem: bufferSize].
		remainingBytesToWrite > 0 ifTrue: [
			self readNextBuffer.
			buffer replaceFrom: 1 to: (position := remainingBytesToWrite) with: aCollection startingAt: startIndex + anInteger - remainingBytesToWrite.
			writeLimit := writeLimit max: position.
			bufferLimit := writeLimit max: readLimit]].
	^aCollection! !

!BufferedReadWriteXtream methodsFor: 'writing' stamp: 'nice 11/28/2009 15:51'!
nextPut: aCharacter

	buffer ifNil: [	self initializeBuffer.].

	position := position + 1.
	position > readLimit ifTrue: [
		"Case we write past end of read buffer"
		position > buffer size ifTrue: [
			self readNextBuffer.
			position := position + 1]].

	writeLimit := writeLimit max: position.
	bufferLimit := writeLimit max: readLimit.
	^buffer at: position put: aCharacter	
	! !

!BufferedReadWriteXtream methodsFor: 'reading' stamp: 'nice 12/3/2009 21:40'!
readInto: aCollection startingAt: startIndex count: n
	"Fill aCollection with next n available characters"

	| nAvailable nToBeRead nRead |
	"check for fast case: enough data in the buffer"
	(nAvailable := self availableDataInBuffer) >= n
		ifTrue: [aCollection replaceFrom: 1 to: n with: buffer startingAt: position + 1.
			position := position + n.
			^n].

	aCollection replaceFrom: 1 to: nAvailable with: buffer startingAt: position + 1.
	(nToBeRead := n - nAvailable) > buffer size
		ifTrue: [
			"since nToBeRead is long, bypass the buffer and read directly from file"
			self flush.
			nRead := source readInto: aCollection startingAt: nAvailable + 1 count: n - nAvailable.
	
			"reset the buffer for next reads"
			position := bufferLimit := readLimit := nRead \\ buffer size.
			buffer replaceFrom: 1 to: position with: aCollection startingAt: 1 + nAvailable + nRead - readLimit]
		ifFalse: [
			self checkAvailableDataInBuffer.
			nRead := nToBeRead min: self availableDataInBuffer.
			aCollection replaceFrom: nAvailable + 1 to: (nAvailable + nRead) with: buffer startingAt: position + 1.
			].
	^nRead + nAvailable! !

!BufferedReadWriteXtream methodsFor: 'private' stamp: 'nice 11/28/2009 16:20'!
readNextBuffer
	"Read a buffer - answer true if some characters are available in the buffer."

	| hasDataAvailable |
	writeLimit > 0 ifTrue: [self flush].
	hasDataAvailable := super readNextBuffer.
	bufferLimit := readLimit.
	^hasDataAvailable! !

!BufferedReadWriteXtream methodsFor: 'private' stamp: 'nice 12/3/2009 23:04'!
writeBufferToFile
	"Write the buffer to file. Leave the file positionned at end of buffer."

	writeLimit > 0 ifFalse: ["No byte were ever written in this buffer - skip the operation"
		^self].
	readLimit > 0 ifTrue: ["Rewind file position to match beginning of buffer"
		source position: source position - readLimit].
	
	"now write the whole buffer"
	"source writeFrom: buffer startingAt: 1 count: (readLimit max: writeLimit)."
	source next: (readLimit max: writeLimit) putAll: buffer startingAt: 1.
	
	"update the readLimit because it should always be in sync with file position"
	bufferLimit := readLimit := readLimit max: writeLimit.
	
	"reset writeLimit to signify that this buffer is clean - that is it has already been written"
	writeLimit := 0! !

!BufferedReadXtream methodsFor: 'file API' stamp: 'nice 12/4/2009 01:35'!
ascii
	source ascii.
	buffer ifNotNil: [buffer := buffer asString].! !

!BufferedReadXtream methodsFor: 'testing' stamp: 'nice 11/28/2009 15:39'!
atEnd
	^self checkAvailableDataInBuffer not! !

!BufferedReadXtream methodsFor: 'private' stamp: 'nice 11/28/2009 15:39'!
availableDataInBuffer
	"Answer the amount of data in buffer than can be read starting from current position."
	
	^readLimit - position! !

!BufferedReadXtream methodsFor: 'file API' stamp: 'nice 12/4/2009 02:50'!
binary
	source binary.
	buffer ifNotNil: [buffer := buffer asByteArray].! !

!BufferedReadXtream methodsFor: 'private' stamp: 'nice 11/28/2009 15:36'!
bufferLimit
	^readLimit! !

!BufferedReadXtream methodsFor: 'converting' stamp: 'nice 12/1/2009 22:02'!
buffered
	^self! !

!BufferedReadXtream methodsFor: 'private' stamp: 'nice 11/28/2009 15:39'!
checkAvailableDataInBuffer
	"Answer true if data is available in buffer.
	In case of end of buffer, force next buffer read."

	readLimit > position ifTrue: [^true].
	^self readNextBuffer! !

!BufferedReadXtream methodsFor: 'private' stamp: 'nice 11/28/2009 15:28'!
clearBuffer
	"Set the internal variables to values signifying there is no data currently in buffer."

	position := readLimit := 0! !

!BufferedReadXtream methodsFor: 'file API' stamp: 'nice 12/4/2009 02:50'!
close
	^source close! !

!BufferedReadXtream methodsFor: 'file API' stamp: 'nice 12/4/2009 02:49'!
closed
	^source closed! !

!BufferedReadXtream methodsFor: 'accessing' stamp: 'nice 12/3/2009 20:58'!
contentsSpecies
	^source contentsSpecies! !

!BufferedReadXtream methodsFor: 'file API' stamp: 'nice 12/4/2009 02:50'!
flush
	^source flush! !

!BufferedReadXtream methodsFor: 'initialize-release' stamp: 'nice 11/28/2009 15:27'!
initialize
	super initialize.
	self clearBuffer! !

!BufferedReadXtream methodsFor: 'private' stamp: 'nice 11/28/2009 16:20'!
initializeBuffer
	buffer := self isBinary
		ifTrue: [ByteArray new: self preferredBufferSize]
		ifFalse: [String new: self preferredBufferSize].
	self clearBuffer! !

!BufferedReadXtream methodsFor: 'testing' stamp: 'nice 12/1/2009 22:00'!
isPositionable
	^true! !

!BufferedReadXtream methodsFor: 'testing' stamp: 'nice 12/3/2009 22:01'!
knowsItsSize
	^source knowsItsSize! !

!BufferedReadXtream methodsFor: 'file API' stamp: 'nice 12/4/2009 02:55'!
name
	^source name! !

!BufferedReadXtream methodsFor: 'reading' stamp: 'nice 12/1/2009 20:22'!
next
	self checkAvailableDataInBuffer ifFalse: [^endOfStreamAction value].
	^buffer at: (position := position + 1)! !

!BufferedReadXtream methodsFor: 'reading' stamp: 'nice 12/1/2009 20:22'!
peek
	self checkAvailableDataInBuffer ifFalse: [^endOfStreamAction value].
	^buffer at: position + 1! !

!BufferedReadXtream methodsFor: 'positionning' stamp: 'nice 11/28/2009 15:52'!
position
	"Answer the read/write position.
	Note that the file position is generally in advance over the read position"
	
	^source position + (position - readLimit)! !

!BufferedReadXtream methodsFor: 'positionning' stamp: 'nice 11/28/2009 15:53'!
position: anInteger
	"move to the indicated position"

	self skip: anInteger - self position! !

!BufferedReadXtream methodsFor: 'private' stamp: 'nice 11/28/2009 15:57'!
preferredBufferSize
	"Answer the preferredSize of the buffer.
	Note: this could be optimized for each OS"
	
	^4096! !

!BufferedReadXtream methodsFor: 'reading' stamp: 'nice 12/3/2009 21:40'!
readInto: aCollection startingAt: startIndex count: n
	"Fill aCollection with next n available characters"

	| nAvailable nToBeRead nRead |
	"check for fast case: enough data in the buffer"
	(nAvailable := self availableDataInBuffer) >= n
		ifTrue: [aCollection replaceFrom: 1 to: n with: buffer startingAt: position + 1.
			position := position + n.
			^n].

	aCollection replaceFrom: 1 to: nAvailable with: buffer startingAt: position + 1.
	(nToBeRead := n - nAvailable) > buffer size
		ifTrue: [
			"since nToBeRead is long, bypass the buffer and read directly from file"
			self flush.
			nRead := source readInto: aCollection startingAt: nAvailable + 1 count: n - nAvailable.
	
			"reset the buffer for next reads"
			position := readLimit := nRead \\ buffer size.
			buffer replaceFrom: 1 to: position with: aCollection startingAt: 1 + nAvailable + nRead - readLimit]
		ifFalse: [
			self checkAvailableDataInBuffer.
			nRead := nToBeRead min: self availableDataInBuffer.
			aCollection replaceFrom: nAvailable + 1 to: (nAvailable + nRead) with: buffer startingAt: position + 1.
			].
	^nRead + nAvailable! !

!BufferedReadXtream methodsFor: 'private' stamp: 'nice 11/28/2009 16:17'!
readNextBuffer
	"Read a buffer - answer true if some characters are available in the buffer."

	buffer ifNil: [self initializeBuffer].
	readLimit < buffer size
		ifTrue: [readLimit := readLimit + (source
			readInto: buffer
			startingAt: readLimit + 1
			count: buffer size - readLimit)]
		ifFalse: [position := 0.
			readLimit := source
				readInto: buffer
				startingAt: 1
				count: buffer size].
	^readLimit > position! !

!BufferedReadXtream methodsFor: 'initialize-release' stamp: 'nice 12/1/2009 22:57'!
reset
	self position: 0! !

!BufferedReadXtream methodsFor: 'positionning' stamp: 'nice 12/3/2009 22:00'!
setToEnd
	(self isPositionable and: [self knowsItsSize])
		ifTrue: [^self position: self size].
	[position := self bufferLimit.
	self checkAvailableDataInBuffer] whileTrue! !

!BufferedReadXtream methodsFor: 'accessing' stamp: 'nice 12/3/2009 22:02'!
size
	self knowsItsSize ifFalse: [self shouldNotImplement].
	^source size! !

!BufferedReadXtream methodsFor: 'positionning' stamp: 'nice 11/28/2009 15:54'!
skip: n
	"Set the character position to n characters from the current position."

	| newBufferPosition newPosition bufferSize |
	newBufferPosition := position + n.
	(newBufferPosition between: 0 and: self bufferLimit)
		ifTrue: ["easy case, new position still point into the buffer"
			position := newBufferPosition.
			^self].
	"hard case, we have to change the buffer"
	self flush.
	newPosition := self position + n max: 0.
	bufferSize := buffer ifNil: [self preferredBufferSize] ifNotNil: [buffer size].
	source position: (newPosition quo: bufferSize) * bufferSize.
	self clearBuffer.
	self readNextBuffer.
	position := (newPosition rem: bufferSize) min: readLimit.! !

!BufferedReadXtream methodsFor: 'reading' stamp: 'nice 11/28/2009 15:37'!
upTo: anObject
	| index result limit |
	self checkAvailableDataInBuffer ifFalse: [^buffer species new].
	
	limit := self bufferLimit.
	index := buffer indexOf: anObject startingAt: position + 1 ifAbsent: [limit + 1].
	"fast case, anObject is in the buffer"
	index <= limit ifTrue: [
		result := buffer copyFrom: position + 1 to: index - 1.
		position := index.
		^result].

	"slow case, repeat the operation over the buffer"
	result := (buffer species new: buffer size * 2) writeXtream.
	[true] whileTrue: [
		result nextPutAll: (buffer copyFrom: position + 1 to: limit).
		position := limit.
		self checkAvailableDataInBuffer ifFalse: [^result contents].
		limit := self bufferLimit.
		index := buffer indexOf: anObject startingAt: position + 1 ifAbsent: [limit + 1].
		index <= limit ifTrue: [
			result nextPutAll: (buffer copyFrom: position + 1 to: index - 1).
			position := index.
			^result contents]]! !

!BufferedReadXtream methodsFor: 'reading' stamp: 'nice 11/28/2009 20:13'!
upToAnyOf: aCollection do: aBlock
	| index result limit |
	self checkAvailableDataInBuffer ifFalse: [^buffer species new].
	limit := self bufferLimit.
	index := buffer indexOfAnyOf: aCollection startingAt: position + 1 ifAbsent: [limit + 1].
	"fast case, aCollection element is in the buffer"
	index <= limit ifTrue: [
		result := buffer copyFrom: position + 1 to: index - 1.
		position := index.
		aBlock value: (buffer at: index).
		^result].

	"slow case, repeat the operation over the buffer"
	result := (buffer species new: buffer size * 2) writeXtream.
	[true] whileTrue: [
		result nextPutAll: (buffer copyFrom: position + 1 to: limit).
		position := limit.
		self checkAvailableDataInBuffer ifFalse: [^result contents].
		limit := self bufferLimit.
		index := buffer indexOfAnyOf: aCollection startingAt: position + 1 ifAbsent: [limit + 1].
		index <= limit ifTrue: [
			result nextPutAll: (buffer copyFrom: position + 1 to: index - 1).
			position := index.
			aBlock value: (buffer at: index).
			^result contents]]! !

!BufferedReadXtream methodsFor: 'reading' stamp: 'nice 12/3/2009 21:51'!
upToEnd
	| result limit |
	self checkAvailableDataInBuffer ifFalse: [^buffer species new].

	result := (buffer species new: buffer size * 2) writeXtream.
	["suppose we don't know the size of the source, just do repeated buffered read"
		limit := self bufferLimit.
		result nextPutAll: (buffer copyFrom: position + 1 to: self bufferLimit).
		position := limit.
		self checkAvailableDataInBuffer] whileTrue.
	^result contents! !

ReadXtream subclass: #CollectionReadXtream
	instanceVariableNames: 'position'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'XTream-Core'!

!CollectionReadXtream commentStamp: 'nice 12/2/2009 04:13' prior: 0!
A CollectionReadXtream is a stream specialized for reading from collections
The collection has to be indexable

Instance Variables
	position:		<Integer> index pointing before the next slot read from collection!

!CollectionReadXtream class methodsFor: 'instance creation' stamp: 'nice 12/2/2009 11:45'!
on: aCollection
	^self new read: aCollection! !

!CollectionReadXtream class methodsFor: 'instance creation' stamp: 'nice 12/2/2009 11:47'!
read: aCollection
	^self new read: aCollection! !

!CollectionReadXtream methodsFor: 'testing' stamp: 'nice 11/28/2009 18:21'!
atEnd
	^position >= source size! !

!CollectionReadXtream methodsFor: 'converting' stamp: 'nice 12/1/2009 23:03'!
buffered
	"No use to buffer a CollectionReadXtream, source already is in memory"
	^self! !

!CollectionReadXtream methodsFor: 'accessing' stamp: 'nice 12/1/2009 21:43'!
contents
	^source! !

!CollectionReadXtream methodsFor: 'accessing' stamp: 'nice 12/1/2009 20:56'!
contentsSpecies
	^source species! !

!CollectionReadXtream methodsFor: 'testing' stamp: 'nice 12/1/2009 21:11'!
isPositionable
	^true! !

!CollectionReadXtream methodsFor: 'testing' stamp: 'nice 12/3/2009 21:56'!
knowsItsSize
	^true! !

!CollectionReadXtream methodsFor: 'reading' stamp: 'nice 11/28/2009 18:04'!
next
	position >= source size ifTrue: [^endOfStreamAction value].
	^source at: (position := 1 + position)! !

!CollectionReadXtream methodsFor: 'reading' stamp: 'nice 11/28/2009 18:32'!
peek
	"faster than super"
	position >= source size ifTrue: [^endOfStreamAction value].
	^source at: 1 + position! !

!CollectionReadXtream methodsFor: 'positionning' stamp: 'nice 11/28/2009 17:59'!
position
	^position! !

!CollectionReadXtream methodsFor: 'positionning' stamp: 'nice 11/28/2009 18:00'!
position: anInteger
	(anInteger between: 0 and: source size) ifFalse: [self error: 'setting stream position out of bounds'].
	position := anInteger! !

!CollectionReadXtream methodsFor: 'initialize-release' stamp: 'nice 11/28/2009 18:24'!
read: aCollection
	source := aCollection.
	position := 0.! !

!CollectionReadXtream methodsFor: 'reading' stamp: 'nice 12/2/2009 00:19'!
readInto: aCollection startingAt: startIndex count: n
	| availableCount |
	availableCount := source size - position min: n.
	aCollection replaceFrom: startIndex to: startIndex + availableCount - 1 with: source startingAt: position.
	^availableCount! !

!CollectionReadXtream methodsFor: 'initialize-release' stamp: 'nice 12/1/2009 22:57'!
reset
	self position: 0! !

!CollectionReadXtream methodsFor: 'positionning' stamp: 'nice 12/1/2009 22:38'!
setToEnd
	self position: self size! !

!CollectionReadXtream methodsFor: 'accessing' stamp: 'nice 12/3/2009 21:57'!
size
	^source size! !

!CollectionReadXtream methodsFor: 'positionning' stamp: 'nice 11/28/2009 18:00'!
skip: anInteger
	self position: position + anInteger! !

!CollectionReadXtream methodsFor: 'reading' stamp: 'nice 12/3/2009 17:16'!
skipThrough: anElement
	position := source indexOf: anElement startingAt: position + 1 ifAbsent: [source size]! !

!CollectionReadXtream methodsFor: 'reading' stamp: 'nice 12/3/2009 21:03'!
skipTo: anElement
	| index |
	index := source indexOf: anElement startingAt: position + 1 ifAbsent: [0].
	position := index = 0
		ifTrue: [source size]
		ifFalse: [index - 1]! !

!CollectionReadXtream methodsFor: 'reading' stamp: 'nice 12/3/2009 17:14'!
through: anElement
	| index result |
	index := source indexOf: anElement startingAt: position + 1 ifAbsent: [0].
	index = 0 ifTrue: [^self upToEnd].
	result := source copyFrom: position + 1 to: index.
	position := index.
	^result! !

!CollectionReadXtream methodsFor: 'reading' stamp: 'nice 12/3/2009 20:18'!
throughAll: subCollection
	"Overriden for speed"

	| index patternSize match i result |
	subCollection isEmpty ifTrue: [^self contentsSpecies new].
	index := position.
	patternSize := subCollection size.
	[
		index := source indexOf: subCollection first startingAt: index + 1 ifAbsent: [0].
		index = 0 ifTrue: [^self upToEnd].
		index := index - 1.
		index + patternSize > source size ifTrue: [^self upToEnd].
		match := true.
		i := 1.
		[match and: [(i := i + 1) <= patternSize]]
			whileTrue: [match := (subCollection at: i) = (source at: index + i)].
		match] whileFalse: [index := index + 1].
	result := source copyFrom: position + 1 to: index + patternSize.
	position := index + patternSize.
	^result! !

!CollectionReadXtream methodsFor: 'reading' stamp: 'nice 12/3/2009 17:14'!
upTo: anElement
	| index result |
	index := source indexOf: anElement startingAt: position + 1 ifAbsent: [0].
	index = 0 ifTrue: [^self upToEnd].
	result := source copyFrom: position + 1 to: index - 1.
	position := index.
	^result! !

!CollectionReadXtream methodsFor: 'reading' stamp: 'nice 12/3/2009 20:18'!
upToAll: subCollection
	"Overriden for speed"

	| index patternSize match i result |
	subCollection isEmpty ifTrue: [^self contentsSpecies new].
	index := position.
	patternSize := subCollection size.
	[
		index := source indexOf: subCollection first startingAt: index + 1 ifAbsent: [0].
		index = 0 ifTrue: [^self upToEnd].
		index := index - 1.
		index + patternSize > source size ifTrue: [^self upToEnd].
		match := true.
		i := 1.
		[match and: [(i := i + 1) <= patternSize]]
			whileTrue: [match := (subCollection at: i) = (source at: index + i)].
		match] whileFalse: [index := index + 1].
	result := source copyFrom: position + 1 to: index.
	position := index + patternSize.
	^result! !

!CollectionReadXtream methodsFor: 'reading' stamp: 'nice 12/3/2009 17:13'!
upToAnyOf: aCollection do: aBlock
	| index result |
	index := source indexOfAnyOf: aCollection startingAt: position + 1 ifAbsent: [0].
	index = 0 ifTrue: [^self upToEnd].
	result := source copyFrom: position + 1 to: index - 1.
	position := index.
	aBlock value: (source at: index).
	^result! !

!CollectionReadXtream methodsFor: 'reading' stamp: 'nice 11/28/2009 20:20'!
upToEnd
	^source copyFrom: position + 1 to: (position := source size).! !

ReadXtream subclass: #FileReadXtream
	instanceVariableNames: 'buffer1'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'XTream-Core'!

!FileReadXtream methodsFor: 'file API' stamp: 'nice 12/1/2009 23:58'!
ascii
	buffer1 := String new: 1! !

!FileReadXtream methodsFor: 'file API' stamp: 'nice 12/1/2009 23:58'!
binary
	buffer1 := ByteArray new: 1! !

!FileReadXtream methodsFor: 'converting' stamp: 'nice 12/1/2009 23:04'!
buffered
	^BufferedReadXtream new source: self! !

!FileReadXtream methodsFor: 'file API' stamp: 'nice 12/4/2009 02:33'!
close
	^source close! !

!FileReadXtream methodsFor: 'file API' stamp: 'nice 12/4/2009 02:33'!
closed
	^source closed! !

!FileReadXtream methodsFor: 'accessing' stamp: 'nice 12/3/2009 23:13'!
contents
	| s savePos |
	savePos := self position.
	self position: 0.
	s := self next: self size.
	self position: savePos.
	^s! !

!FileReadXtream methodsFor: 'accessing' stamp: 'nice 12/4/2009 02:34'!
contentsSpecies
	^buffer1 class! !

!FileReadXtream methodsFor: 'file API' stamp: 'nice 12/4/2009 02:43'!
flush
	^source flush! !

!FileReadXtream methodsFor: 'initialize-release' stamp: 'nice 12/4/2009 02:35'!
initialize
	self binary! !

!FileReadXtream methodsFor: 'testing' stamp: 'nice 12/3/2009 21:18'!
isPositionable
	^true! !

!FileReadXtream methodsFor: 'testing' stamp: 'nice 12/3/2009 21:59'!
knowsItsSize
	^true! !

!FileReadXtream methodsFor: 'file API' stamp: 'nice 12/3/2009 23:14'!
name
	^source name! !

!FileReadXtream methodsFor: 'reading' stamp: 'nice 12/1/2009 23:52'!
next
	(source next: 1 into: buffer1 startingAt: 1) = 1 ifFalse: [^endOfStreamAction value].
	^buffer1 first! !

!FileReadXtream methodsFor: 'positionning' stamp: 'nice 12/3/2009 21:19'!
position
	^source position! !

!FileReadXtream methodsFor: 'positionning' stamp: 'nice 12/3/2009 21:27'!
position: anInteger
	^source position: anInteger! !

!FileReadXtream methodsFor: 'reading' stamp: 'nice 12/2/2009 00:19'!
readInto: aCollection startingAt: startIndex count: anInteger
	^source readInto: aCollection startingAt: startIndex count: anInteger! !

!FileReadXtream methodsFor: 'accessing' stamp: 'nice 12/3/2009 21:59'!
size
	^source size! !

!FileReadXtream methodsFor: 'positionning' stamp: 'nice 12/3/2009 21:27'!
skip: anInteger
	^source position: source position + anInteger! !

!ReadXtream methodsFor: 'reading' stamp: 'nice 12/1/2009 22:53'!
back
	"Go back one element and return it."
	self position = 0 ifTrue: [self errorCantGoBack].
	self skip: -1.
	^ self peek! !

!ReadXtream methodsFor: 'deprecated' stamp: 'nice 12/2/2009 00:10'!
basicNext
	^self next! !

!ReadXtream methodsFor: 'enumerating' stamp: 'nice 12/3/2009 10:09'!
collect: aBlock
	^self collect: aBlock 	as: self contentsSpecies! !

!ReadXtream methodsFor: 'enumerating' stamp: 'nice 12/4/2009 02:04'!
collect: aBlock as: collectionSpecies
	| endofStreamMark |
	endofStreamMark := self goodEndOfStreamMark.
	^self
		transform: [
			| element |
			(element := self next) = endofStreamMark
				ifTrue: [element]
				ifFalse: [aBlock value: element]]
		as: collectionSpecies
		endOfStreamValue: endofStreamMark! !

!ReadXtream methodsFor: 'accessing' stamp: 'nice 12/3/2009 10:18'!
contentsSpecies
	"indicate a collection species suitable for reading a subcollection of this stream"
	self subclassResponsibility! !

!ReadXtream methodsFor: 'converting' stamp: 'nice 12/4/2009 02:44'!
decodeWith: aTextConverter
	self ascii.	"TextConverter would not do anything of a ByteArray... That's silly"
	^self transform: [aTextConverter nextFromStream: self] as: String! !

!ReadXtream methodsFor: 'enumerating' stamp: 'nice 12/3/2009 09:42'!
do: aBlock
	"Evaluate aBlock with each element found on this stream.
	Return self when no more data is available, do not execute endOfStreamAction"

	self setEndOfStreamAction: [^self] while:
		[[true] whileTrue: [aBlock value: self next]]! !

!ReadXtream methodsFor: 'accessing' stamp: 'nice 12/2/2009 01:35'!
endOfStreamAction
	^endOfStreamAction! !

!ReadXtream methodsFor: 'accessing' stamp: 'nice 12/3/2009 09:52'!
endOfStreamAction: aValuable
	"Set the action to evaluate when end of stream is encountered.
	This action must understand the message #value.
	Answer previous value of endOfStreamAction"
	| old |
	old := endOfStreamAction.
	endOfStreamAction := aValuable.
	^old! !

!ReadXtream methodsFor: 'private' stamp: 'nice 12/4/2009 02:05'!
goodEndOfStreamMark
	"Answer a good end of stream mark, that is an object (self next) can asnwser to indicate self atEnd."
	
	^self contentsSpecies isPointers
		ifTrue: ["nil is not a good endOfStreamMark because I can contain nil. Use an Object"
			Object new]
		ifFalse: [nil].! !

!ReadXtream methodsFor: 'file API' stamp: 'nice 12/4/2009 02:38'!
isBinary
	^self contentsSpecies = ByteArray! !

!ReadXtream methodsFor: 'testing' stamp: 'nice 12/1/2009 21:03'!
isReadable
	^true! !

!ReadXtream methodsFor: 'testing' stamp: 'nice 12/3/2009 21:56'!
knowsItsSize
	"Answer true if the receiver knows its size.
	Default is false. If true, subclass should override and also define #size."
	
	^false! !

!ReadXtream methodsFor: 'reading' stamp: 'nice 12/3/2009 09:42'!
match: subCollection
	"Set the access position of the receiver to be past the next occurrence of the subCollection. Answer whether subCollection is found.  No wildcards, and case does matter."

	^self restoreEndOfStreamActionAfter:
		[| pattern startMatch |
		pattern := subCollection readXtream.
		pattern endOfStreamAction: [^true].
		self endOfStreamAction: [^false].
		startMatch := nil.
		[true] whileTrue: [pattern next = self next
			ifTrue: [pattern position = 1 ifTrue: [startMatch := self position]]
			ifFalse: [pattern position: 0.
				startMatch ifNotNil: [
					self position: startMatch.
					startMatch := nil]]]]! !

!ReadXtream methodsFor: 'reading' stamp: 'nice 12/1/2009 23:14'!
next
	self subclassResponsibility! !

!ReadXtream methodsFor: 'reading' stamp: 'nice 12/2/2009 00:21'!
next: count
	^self next: count into: (self contentsSpecies ofSize: count) startingAt: 1! !

!ReadXtream methodsFor: 'reading' stamp: 'nice 12/2/2009 00:21'!
next: count into: aCollection
	^self next: count into: aCollection startingAt: 1! !

!ReadXtream methodsFor: 'reading' stamp: 'nice 12/2/2009 00:21'!
next: count into: aCollection startingAt: startIndex
	| readCount |
	readCount := self readInto: aCollection startingAt: startIndex count: count.
	readCount = count
		ifFalse: [endOfStreamAction value.
			self error: 'incomplete count read from stream'].
	^aCollection! !

!ReadXtream methodsFor: 'reading' stamp: 'nice 12/2/2009 00:56'!
nextAvailable: count
	| aCollection readCount |
	aCollection := self contentsSpecies ofSize: count.
	readCount := self next: count into: aCollection startingAt: 1.
	^readCount = count
		ifTrue: [aCollection]
		ifFalse: [aCollection copyFrom: 1 to: readCount]! !

!ReadXtream methodsFor: 'reading' stamp: 'nice 12/3/2009 09:42'!
nextDelimited: terminator
	"Answer the contents of the receiver, up to the next terminator character. Doubled terminators indicate an embedded terminator character.  For example: 'this '' was a quote'. Start postioned before the initial terminator."

	^self restoreEndOfStreamActionAfter:
		[| newXtream element |
		newXtream := (self contentsSpecies new: 100) writeXtream.
		self endOfStreamAction: [^newXtream contents].
		self peekFor: terminator.	"absorb initial terminator"
		[(element := self next) = terminator and: [(self peekFor: terminator) not]] whileFalse: [ newXtream nextPut: element].
		newXtream contents]! !

!ReadXtream methodsFor: 'reading' stamp: 'nice 12/2/2009 00:21'!
nextInto: aCollection
	^self next: aCollection size into: aCollection startingAt: 1! !

!ReadXtream methodsFor: 'ascii API' stamp: 'nice 12/1/2009 23:15'!
nextLine
	"Answer the next line up to any of CR, LF or CR LF pair, non inclusive"
	
	^self upToAnyOf: CharacterSet crlf do: [:char | char = Character cr ifTrue: [self peekFor: Character lf]].! !

!ReadXtream methodsFor: 'reading' stamp: 'nice 12/1/2009 22:54'!
oldBack
	"Go back one element and return it.  Use indirect messages in case I am a StandardFileXtream"
	"The method is a misconception about what a stream is. A stream contains a pointer *between* elements with past and future elements. This method considers that the pointer is *on* an element. Please consider unit tests which verifies #back and #oldBack behavior. (Damien Cassou - 1 August 2007)"
	self position = 0 ifTrue: [self errorCantGoBack].
	self position = 1 ifTrue: [self position: 0.  ^ endOfStreamAction value].
	self skip: -2.
	^ self next! !

!ReadXtream methodsFor: 'reading' stamp: 'nice 11/28/2009 18:31'!
peek
	| anObject |
	self atEnd ifTrue: [^endOfStreamAction value].
	anObject := self next.
	self skip: -1.
	^anObject! !

!ReadXtream methodsFor: 'reading' stamp: 'nice 11/28/2009 18:33'!
peekFor: anObject
	self atEnd ifTrue: [^endOfStreamAction value].
	self next = anObject ifTrue: [^true].
	self skip: -1.
	^false! !

!ReadXtream methodsFor: 'reading' stamp: 'nice 12/3/2009 21:41'!
readInto: aCollection startingAt: startIndex count: anInteger
	"Attempt to read anInteger element into aCollection starting at index startIndex.
	Return the count effectively read.
	Do not evaluate endOfStreamAction if no more data available"
	^self subclassResponsibility! !

!ReadXtream methodsFor: 'enumerating' stamp: 'nice 12/4/2009 02:04'!
reject: aBlock
	| endofStreamMark |
	endofStreamMark := self goodEndOfStreamMark.
	^self
		transform: [
			| element |
			[(element := self next) ~= endofStreamMark and: [aBlock value: element]] whileTrue.
			element]
		as: self contentsSpecies
		endOfStreamValue: endofStreamMark! !

!ReadXtream methodsFor: 'private' stamp: 'nice 12/2/2009 01:32'!
restoreEndOfStreamActionAfter: aBlock
	| restored |
	restored := endOfStreamAction.
	^[aBlock value] ensure: [self endOfStreamAction: restored]! !

!ReadXtream methodsFor: 'enumerating' stamp: 'nice 12/4/2009 02:04'!
select: aBlock
	| endofStreamMark |
	endofStreamMark := self goodEndOfStreamMark.
	^self
		transform: [
			"Make a block that will answer next element of self satisying aBlock or endOfStreamMark if none"
			| element |
			[(element := self next) = endofStreamMark or: [aBlock value: element]] whileFalse.
			element]
		as: self contentsSpecies
		endOfStreamValue: endofStreamMark! !

!ReadXtream methodsFor: 'private' stamp: 'nice 12/2/2009 01:33'!
setEndOfStreamAction: action while: aBlock
	| restored |
	restored := endOfStreamAction.
	^
	[self endOfStreamAction: action.
	aBlock value] ensure: [self endOfStreamAction: restored]! !

!ReadXtream methodsFor: 'reading' stamp: 'nice 12/3/2009 21:58'!
setToEnd
	"Position the stream at end of available data"

	(self isPositionable and: [self knowsItsSize])
		ifTrue: [^self position: self size].
	self setEndOfStreamAction: [^self] while:
		[[true] whileTrue: [self next]]! !

!ReadXtream methodsFor: 'accessing' stamp: 'nice 12/3/2009 21:56'!
size
	self knowsItsSize
		ifTrue: [self subclassResponsibility]
		ifFalse: [self shouldNotImplement]! !

!ReadXtream methodsFor: 'positionning' stamp: 'nice 12/3/2009 14:59'!
skip: anInteger
	"provide a default slow and naive implementation for forward skip"
	anInteger < 0 ifTrue: [
		self isPositionable
			ifTrue: [self subclassResponsibility]
			ifFalse: [self errorNotPositionable]].
	1 to: anInteger do: [:i | self next]! !

!ReadXtream methodsFor: 'reading' stamp: 'nice 12/3/2009 09:42'!
skipThrough: anObject 
	"position the stream after the next occurence of anObject"

	self restoreEndOfStreamActionAfter:
		[self endOfStreamAction: [anObject].
		[self next = anObject] whileFalse]! !

!ReadXtream methodsFor: 'reading' stamp: 'nice 12/3/2009 09:42'!
skipTo: anObject 
	"position the stream before the next occurence of anObject"

	self setEndOfStreamAction: [^self] while:
		[[self next = anObject] whileFalse.
		self skip: -1]! !

!ReadXtream methodsFor: 'accessing' stamp: 'nice 12/1/2009 22:34'!
source: anObject
	"Set the source of this stream"
	
	source := anObject! !

!ReadXtream methodsFor: 'reading' stamp: 'nice 12/3/2009 09:42'!
through: anObject 
	"Answer a subcollection from the current access position to the 
	occurrence (if any, inclusive) of anObject in the receiver. If 
	anObject is not in the collection, answer the entire rest of the receiver."

	^self restoreEndOfStreamActionAfter:
		[| newXtream element |
		newXtream := (self contentsSpecies new: 100) writeXtream.
		self endOfStreamAction: [^newXtream contents].
		[ newXtream nextPut: (element := self next).
		element = anObject] whileFalse.
		newXtream contents]! !

!ReadXtream methodsFor: 'reading' stamp: 'nice 12/3/2009 09:42'!
throughAll: subCollection
	"Answer a subcollection from the current access position to the occurrence (if any, inclusive) of aCollection. If aCollection is not in the stream, answer the entire rest of the stream."

	^self restoreEndOfStreamActionAfter:
		[| pattern startMatch result element |
		result := (self contentsSpecies ofSize: 100) writeXtream.
		pattern := subCollection readXtream.
		pattern endOfStreamAction: [
			^result nextPutAll: subCollection; contents].
		self endOfStreamAction: [^result contents].
		startMatch := nil.
		[true] whileTrue: [pattern next = (element := self next)
			ifTrue: [pattern position = 1 ifTrue: [startMatch := self position]]
			ifFalse: [
				startMatch
					ifNil: [result nextPut: element]
					ifNotNil: ["restart one character after the first match"
						result nextPut: subCollection first.
						self position: startMatch.
						startMatch := nil].
				pattern position: 0]]]! !

!ReadXtream methodsFor: 'enumerating' stamp: 'nice 12/4/2009 02:09'!
transform: aBlock as: collectionSpecies
	^self transform: aBlock as: collectionSpecies endOfStreamValue: self goodEndOfStreamMark! !

!ReadXtream methodsFor: 'enumerating' stamp: 'nice 12/4/2009 03:01'!
transform: aBlock as: collectionSpecies endOfStreamValue: anObject
	^(BlockReadXtreamWrapper new)
		source: aBlock;
		origin: self;
		contentsSpecies: collectionSpecies;
		endOfStreamAction: (self endOfStreamAction: nil->anObject);
		endOfStreamValue: anObject;
		yourself! !

!ReadXtream methodsFor: 'reading' stamp: 'nice 12/3/2009 09:42'!
upTo: anObject 
	"Answer a subcollection from the current access position to the 
	occurrence (if any, but not inclusive) of anObject in the receiver. If 
	anObject is not in the collection, answer the entire rest of the receiver."

	^self restoreEndOfStreamActionAfter:
		[| newXtream element |
		newXtream := (self contentsSpecies new: 100) writeXtream.
		self endOfStreamAction: [anObject].
		[(element := self next) = anObject] whileFalse: [ newXtream nextPut: element].
		newXtream contents]! !

!ReadXtream methodsFor: 'reading' stamp: 'nice 12/3/2009 09:42'!
upToAll: subCollection
	"Answer a subcollection from the current access position to the occurrence (if any, but not inclusive) of aCollection. If aCollection is not in the stream, answer the entire rest of the stream."

	^self restoreEndOfStreamActionAfter:
		[| pattern startMatch result element |
		result := (self contentsSpecies ofSize: 100) writeXtream.
		pattern := subCollection readXtream.
		pattern endOfStreamAction: [^result contents].
		self endOfStreamAction: [^result contents].
		startMatch := nil.
		[true] whileTrue: [pattern next = (element := self next)
			ifTrue: [startMatch ifNil: [startMatch := self position]]
			ifFalse: [
				startMatch
					ifNil: [result nextPut: element]
					ifNotNil: ["restart one character after the first match"
						result nextPut: subCollection first.
						self position: startMatch.
						startMatch := nil].
				pattern position: 0]]]! !

!ReadXtream methodsFor: 'reading' stamp: 'nice 12/1/2009 20:22'!
upToAnyOf: aCollection 
	"Answer a subcollection from the current access position to the 
	occurrence (if any, but not inclusive) of anObject in the receiver. If 
	anObject is not in the collection, answer the entire rest of the receiver."

	^self upToAnyOf: aCollection do: [:stopObject | ]! !

!ReadXtream methodsFor: 'reading' stamp: 'nice 12/3/2009 09:42'!
upToAnyOf: aCollection do: aBlock
	"Answer a subcollection from the current access position to the 
	occurrence (if any, but not inclusive) of anObject in the receiver. If 
	anObject is not in the collection, answer the entire rest of the receiver.
	Evaluate aBlock with the first encountered object in aCollection"

	aCollection ifEmpty: [^self upToEnd].
	^self restoreEndOfStreamActionAfter:
		[| newXtream element |
		newXtream := (self contentsSpecies new: 100) writeXtream.
		self endOfStreamAction: [aCollection anyOne].
		[aCollection includes: (element := self next)] whileFalse: [ newXtream nextPut: element ].
		aBlock value: element.
		newXtream contents]! !

!ReadXtream methodsFor: 'reading' stamp: 'nice 12/3/2009 14:45'!
upToEnd
	"Answer a subcollection from the current access position through the last element of the receiver."

	^self restoreEndOfStreamActionAfter:
		[| newXtream |
		newXtream := (self contentsSpecies new: 100) writeXtream.
		self endOfStreamAction: [^newXtream contents].
		[true] whileTrue: [ newXtream nextPut: self next ]]! !

Xtream subclass: #WriteXtream
	instanceVariableNames: 'destination writeLimit'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'XTream-Core'!

!WriteXtream commentStamp: 'nice 12/2/2009 04:00' prior: 0!
A WriteXtream is an abstract class for writing a sequence of objects.

Instance Variables
	destination:		<Object>
	writeLimit:		<Integer>

destination
	- the handle for storing the sequence of objects

writeLimit
	- xxxxx
!

WriteXtream subclass: #BlockWriteXtream
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'XTream-Core'!

!BlockWriteXtream commentStamp: 'nice 12/3/2009 14:22' prior: 0!
A BlockWriteXtream is a stream using a one argument block as its destination
!

!BlockWriteXtream methodsFor: 'accessing' stamp: 'nice 12/3/2009 14:33'!
contents
	"BlockWriteXtream do not remember their contents.
	They just write and forget"
	self error: 'BlockWriteXtream do not remember their contents'! !

!BlockWriteXtream methodsFor: 'writing' stamp: 'nice 12/3/2009 14:49'!
next: count putAll: aCollection startingAt: startIndex 
	| index nWrite |
	index := startIndex.
	nWrite := 0.
	[nWrite < count] whileTrue: [
		destination value: (aCollection at: index).
		index := index + 1.
		nWrite := nWrite + 1].
	^aCollection! !

!BlockWriteXtream methodsFor: 'writing' stamp: 'nice 12/3/2009 14:22'!
nextPut: anObject
	destination value: anObject.
	^anObject! !

BlockWriteXtream subclass: #BlockWriteXtreamWrapper
	instanceVariableNames: 'target'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'XTream-Core'!

!BlockWriteXtreamWrapper commentStamp: 'nice 12/3/2009 14:52' prior: 0!
A BlockWriteXtreamWrapper is a stream using a one argument block as its destination
It also has a pointer to the wrapped stream so as to be able to delegate some message to it.

Instance Variables
	target	<WriteXtream> the destination stream!

!BlockWriteXtreamWrapper methodsFor: 'file API' stamp: 'nice 12/3/2009 14:25'!
close
	^target close! !

!BlockWriteXtreamWrapper methodsFor: 'file API' stamp: 'nice 12/3/2009 14:24'!
closed
	^target closed! !

!BlockWriteXtreamWrapper methodsFor: 'accessing' stamp: 'nice 12/3/2009 10:17'!
contents
	^target contents! !

!BlockWriteXtreamWrapper methodsFor: 'accessing' stamp: 'nice 12/3/2009 14:53'!
contentsSpecies
	^target contentsSpecies! !

!BlockWriteXtreamWrapper methodsFor: 'file API' stamp: 'nice 12/3/2009 14:25'!
flush
	^target flush! !

!BlockWriteXtreamWrapper methodsFor: 'file API' stamp: 'nice 12/4/2009 02:54'!
name
	^target name! !

!BlockWriteXtreamWrapper methodsFor: 'accessing' stamp: 'nice 12/3/2009 14:20'!
target: aWriteXtream
	^target := aWriteXtream! !

WriteXtream subclass: #CollectionWriteXtream
	instanceVariableNames: 'position'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'XTream-Core'!

!CollectionWriteXtream commentStamp: 'nice 12/2/2009 04:12' prior: 0!
A CollectionWriteXtream is stream specialized for writing into collections
The collection has to be indexable

Instance Variables
	position:		<Integer> index pointing before the next written slot in destination collection!

!CollectionWriteXtream class methodsFor: 'instance creation' stamp: 'nice 12/2/2009 11:47'!
append: aCollection
	^self new append: aCollection! !

!CollectionWriteXtream class methodsFor: 'instance creation' stamp: 'nice 12/2/2009 11:47'!
on: aCollection
	^self new write: aCollection! !

!CollectionWriteXtream class methodsFor: 'instance creation' stamp: 'nice 12/2/2009 11:47'!
write: aCollection
	^self new write: aCollection! !

!CollectionWriteXtream methodsFor: 'initialize-release' stamp: 'nice 12/1/2009 22:06'!
append: aCollection
	destination := aCollection.
	position := writeLimit := destination size.! !

!CollectionWriteXtream methodsFor: 'accessing' stamp: 'nice 12/1/2009 20:59'!
contents
	^destination copyFrom: 1 to: position! !

!CollectionWriteXtream methodsFor: 'accessing' stamp: 'nice 12/1/2009 20:56'!
contentsSpecies
	^destination species! !

!CollectionWriteXtream methodsFor: 'file API' stamp: 'nice 12/4/2009 02:39'!
isBinary
	^destination class = ByteArray! !

!CollectionWriteXtream methodsFor: 'testing' stamp: 'nice 12/1/2009 21:11'!
isPositionable
	^true! !

!CollectionWriteXtream methodsFor: 'writing' stamp: 'nice 12/2/2009 12:28'!
next: count putAll: aCollection startingAt: startIndex
	"Grow the destination by doubling the size, but keeping the growth between 20 and 1000000.
	Then put <anObject> at the current write position."
	
	| lastPosition growSize |
	(lastPosition := position + count) > destination size ifTrue: [
		growSize := ((destination size max: 20) min: 1000000) max: lastPosition - destination size.
		destination := destination grownBy: growSize].
	destination replaceFrom: position + 1 to: lastPosition with: aCollection startingAt: startIndex.
	position := lastPosition.
	writeLimit := writeLimit max: position.
	^aCollection! !

!CollectionWriteXtream methodsFor: 'writing' stamp: 'nice 12/1/2009 22:09'!
nextPut: anObject
	| growSize |
	position >= destination size ifTrue: [
		growSize := (destination size max: 20) min: 1000000.
		destination := destination grownBy: growSize].
	destination at: (position := 1 + position) put: anObject.
	writeLimit := writeLimit max: position.
	^anObject! !

!CollectionWriteXtream methodsFor: 'positionning' stamp: 'nice 11/28/2009 17:59'!
position
	^position! !

!CollectionWriteXtream methodsFor: 'positionning' stamp: 'nice 12/1/2009 22:06'!
position: anInteger
	(anInteger between: 0 and: writeLimit) ifFalse: [self error: 'setting stream position out of bounds'].
	position := anInteger! !

!CollectionWriteXtream methodsFor: 'initialize-release' stamp: 'nice 12/1/2009 22:58'!
reset
	self position: 0! !

!CollectionWriteXtream methodsFor: 'positionning' stamp: 'nice 12/1/2009 22:39'!
setToEnd
	self position: writeLimit! !

!CollectionWriteXtream methodsFor: 'accessing' stamp: 'nice 12/1/2009 22:38'!
size
	^writeLimit! !

!CollectionWriteXtream methodsFor: 'positionning' stamp: 'nice 11/28/2009 18:00'!
skip: anInteger
	self position: position + anInteger! !

!CollectionWriteXtream methodsFor: 'initialize-release' stamp: 'nice 12/2/2009 12:01'!
write: aCollection
	destination := aCollection.
	position := writeLimit := 0.! !

WriteXtream subclass: #FileWriteXtream
	instanceVariableNames: 'buffer1'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'XTream-Core'!

!FileWriteXtream methodsFor: 'initialize-release' stamp: 'nice 12/3/2009 23:30'!
append: aFile
	destination := aFile.
	destination setToEnd! !

!FileWriteXtream methodsFor: 'file API' stamp: 'nice 12/1/2009 23:58'!
ascii
	buffer1 := String new: 1! !

!FileWriteXtream methodsFor: 'file API' stamp: 'nice 12/1/2009 23:58'!
binary
	buffer1 := ByteArray new: 1! !

!FileWriteXtream methodsFor: 'converting' stamp: 'nice 12/2/2009 00:05'!
buffered
	^BufferedReadWriteXtream new source: self! !

!FileWriteXtream methodsFor: 'file API' stamp: 'nice 12/4/2009 02:36'!
close
	^destination close! !

!FileWriteXtream methodsFor: 'file API' stamp: 'nice 12/4/2009 02:36'!
closed
	^destination closed! !

!FileWriteXtream methodsFor: 'accessing' stamp: 'nice 12/4/2009 02:47'!
contentsSpecies
	^buffer1 class! !

!FileWriteXtream methodsFor: 'file API' stamp: 'nice 12/4/2009 02:42'!
flush
	^destination flush! !

!FileWriteXtream methodsFor: 'initialize-release' stamp: 'nice 12/4/2009 02:35'!
initialize
	self binary! !

!FileWriteXtream methodsFor: 'file API' stamp: 'nice 12/4/2009 02:40'!
isBinary
	^buffer1 class = ByteArray! !

!FileWriteXtream methodsFor: 'testing' stamp: 'nice 12/3/2009 21:18'!
isPositionable
	^true! !

!FileWriteXtream methodsFor: 'file API' stamp: 'nice 12/3/2009 23:14'!
name
	^destination name! !

!FileWriteXtream methodsFor: 'writing' stamp: 'nice 12/1/2009 23:54'!
next: count putAll: aCollection startingAt: startIndex
	^destination next: count putAll: aCollection startingAt: startIndex! !

!FileWriteXtream methodsFor: 'writing' stamp: 'nice 12/3/2009 08:37'!
nextPut: anObject
	buffer1 at: 1 put: anObject.
	destination next: 1 putAll: buffer1 startingAt: 1.
	^anObject! !

!FileWriteXtream methodsFor: 'positionning' stamp: 'nice 12/3/2009 21:27'!
position
	^destination position! !

!FileWriteXtream methodsFor: 'positionning' stamp: 'nice 12/3/2009 21:27'!
position: anInteger
	^destination position: anInteger! !

!FileWriteXtream methodsFor: 'reading' stamp: 'nice 12/3/2009 23:09'!
readInto: aCollection startingAt: startIndex count: anInteger
	"Assume files are read/write.
	This is used by buffers to read back previous contents when positionning back"
	
	^destination readInto: aCollection startingAt: startIndex count: anInteger! !

!FileWriteXtream methodsFor: 'positionning' stamp: 'nice 12/3/2009 21:28'!
skip: anInteger
	^destination position: destination position + anInteger! !

!FileWriteXtream methodsFor: 'initialize-release' stamp: 'nice 12/3/2009 23:31'!
write: aFile
	destination := aFile! !

!WriteXtream methodsFor: 'deprecated' stamp: 'nice 12/2/2009 00:10'!
basicNextPut: aChar
	^self nextPut: aChar! !

!WriteXtream methodsFor: 'deprecated' stamp: 'nice 12/2/2009 00:11'!
basicNextPutAll: aCollection
	^self nextPutAll: aCollection! !

!WriteXtream methodsFor: 'enumerating' stamp: 'nice 12/3/2009 14:54'!
collect: aBlock
	^self
		transform: [:anObject |
			"Make a block that will nextPut: element of self transformed by aBlock"
			self nextPut: (aBlock value: anObject)]! !

!WriteXtream methodsFor: 'ascii API' stamp: 'nice 11/28/2009 18:28'!
cr
	self nextPut: Character cr! !

!WriteXtream methodsFor: 'ascii API' stamp: 'nice 11/28/2009 18:29'!
crlf
	self cr; lf! !

!WriteXtream methodsFor: 'ascii API' stamp: 'nice 11/28/2009 18:29'!
crtab
	self cr; tab! !

!WriteXtream methodsFor: 'ascii API' stamp: 'nice 11/28/2009 18:36'!
crtab: anInteger 
	"Append a carriage return and anInteger tab characters to the receiver."

	self cr; tab: anInteger! !

!WriteXtream methodsFor: 'accessing' stamp: 'nice 12/1/2009 22:35'!
destination: anObject
	"Set the destination of this stream"
	
	destination := anObject! !

!WriteXtream methodsFor: 'converting' stamp: 'nice 12/4/2009 01:52'!
encodeWith: aTextConverter
	self ascii.	"TextConverter would not do anything of a ByteArray... That's silly"
	^self transform: [:e | aTextConverter nextPut: e toStream: self. e]! !

!WriteXtream methodsFor: 'testing' stamp: 'nice 12/1/2009 21:03'!
isWriteable
	^true! !

!WriteXtream methodsFor: 'ascii API' stamp: 'nice 11/28/2009 18:28'!
lf
	self nextPut: Character lf! !

!WriteXtream methodsFor: 'writing' stamp: 'nice 12/2/2009 02:15'!
next: count putAll: aCollection 
	^self next: count putAll: aCollection startingAt: 1! !

!WriteXtream methodsFor: 'writing' stamp: 'nice 12/1/2009 23:17'!
next: count putAll: aCollection startingAt: startIndex
	self subclassResponsibility! !

!WriteXtream methodsFor: 'writing' stamp: 'nice 12/1/2009 23:17'!
nextPut: anObject
	self subclassResponsibility! !

!WriteXtream methodsFor: 'writing' stamp: 'nice 11/28/2009 17:49'!
nextPutAll: aCollection
	self next: aCollection size putAll: aCollection startingAt: 1! !

!WriteXtream methodsFor: 'writing' stamp: 'nice 11/28/2009 18:35'!
print: anObject
	anObject printOn: self! !

!WriteXtream methodsFor: 'enumerating' stamp: 'nice 12/3/2009 14:54'!
reject: aBlock
	^self
		transform: [:anObject |
			"Make a block that will nextPut: element of self not satisying aBlock"
			(aBlock value: anObject) ifFalse: [self nextPut: anObject].
			anObject]! !

!WriteXtream methodsFor: 'enumerating' stamp: 'nice 12/3/2009 14:54'!
select: aBlock
	^self
		transform: [:anObject |
			"Make a block that will nextPut: element of self satisying aBlock"
			(aBlock value: anObject) ifTrue: [self nextPut: anObject].
			anObject]! !

!WriteXtream methodsFor: 'ascii API' stamp: 'nice 11/28/2009 18:28'!
space
	self nextPut: Character space! !

!WriteXtream methodsFor: 'ascii API' stamp: 'nice 11/28/2009 18:29'!
space: anInteger 
	"Append anInteger space characters to the receiver."

	anInteger timesRepeat: [self space]! !

!WriteXtream methodsFor: 'writing' stamp: 'nice 11/28/2009 18:35'!
store: anObject
	anObject storeOn: self! !

!WriteXtream methodsFor: 'ascii API' stamp: 'nice 11/28/2009 18:28'!
tab
	self nextPut: Character tab! !

!WriteXtream methodsFor: 'ascii API' stamp: 'nice 11/28/2009 18:29'!
tab: anInteger 
	"Append anInteger tab characters to the receiver."

	anInteger timesRepeat: [self tab]! !

!WriteXtream methodsFor: 'enumerating' stamp: 'nice 12/3/2009 14:27'!
transform: aBlock 
	^BlockWriteXtreamWrapper new
		destination: aBlock;
		target: self;
		yourself! !

!Xtream methodsFor: 'file API' stamp: 'nice 12/1/2009 21:17'!
ascii
	"This is part of external stream API, indicating that character input/output is requested.
	Squeak tradition is to silently ignore such message in any other stream."
	
	InappropriateXtreamMessage signal: thisContext selector.
	^self! !

!Xtream methodsFor: 'testing' stamp: 'nice 12/4/2009 02:15'!
atEnd
	"Tells whether this stream is atEnd, which means no data is available for read"
	
	self isReadable
		ifTrue: [self subclassResponsibility]
		ifFalse: [self errorNotReadable]! !

!Xtream methodsFor: 'file API' stamp: 'nice 12/1/2009 21:17'!
binary
	"This is part of external stream API, indicating that raw bytes input/output is requested.
	Squeak tradition is to silently ignore such message in any other stream."
	
	InappropriateXtreamMessage signal: thisContext selector.
	^self! !

!Xtream methodsFor: 'file API' stamp: 'nice 12/1/2009 21:17'!
close
	"This is part of external stream API, indicating that we wish to close the external stream.
	Squeak tradition is to silently ignore such message in any other stream."
	
	InappropriateXtreamMessage signal: thisContext selector.
	^self! !

!Xtream methodsFor: 'file API' stamp: 'nice 12/1/2009 21:18'!
closed
	"This is part of external stream API, inquiring whether the external stream is closed or not.
	Squeak tradition is to silently ignore such message in any other stream."
	
	InappropriateXtreamMessage signal: thisContext selector.
	^false! !

!Xtream methodsFor: 'accessing' stamp: 'nice 12/1/2009 21:43'!
contents
	self subclassResponsibility! !

!Xtream methodsFor: 'accessing' stamp: 'nice 12/1/2009 21:08'!
contentsSpecies
	"indicate a collection species suitable for reading a subcollection of this stream"
	self subclassResponsibility! !

!Xtream methodsFor: 'errors' stamp: 'nice 12/1/2009 22:55'!
errorCantGoBack
	^self error: 'This stream is at beginning and cannot go back'! !

!Xtream methodsFor: 'errors' stamp: 'nice 12/1/2009 21:21'!
errorNotPositionable
	^self error: 'This stream is not positionable'! !

!Xtream methodsFor: 'errors' stamp: 'nice 12/1/2009 21:20'!
errorNotReadable
	^self error: 'This stream is not readable'! !

!Xtream methodsFor: 'errors' stamp: 'nice 12/1/2009 21:20'!
errorNotWriteable
	^self error: 'This stream is not writeable'! !

!Xtream methodsFor: 'file API' stamp: 'nice 12/1/2009 21:58'!
flush
	"This is part of external stream API, indicating that we wish to force writing of data to the external stream.
	Squeak tradition is to silently ignore such message in any other stream."
	
	^self! !

!Xtream methodsFor: 'file API' stamp: 'nice 12/1/2009 21:51'!
isBinary
	"This is part of external stream API, inquiring whether the external stream wll answer integer or characters.
	Squeak tradition is to silently ignore such message in any other stream."
	
	InappropriateXtreamMessage signal: thisContext selector.
	^false! !

!Xtream methodsFor: 'testing' stamp: 'nice 12/1/2009 21:04'!
isPositionable
	"Tells whether this stream respondsTo: position, position:, skip: protocol.
	Default is false, subclass should redefine if true."
	^false! !

!Xtream methodsFor: 'testing' stamp: 'nice 12/1/2009 21:05'!
isReadable
	"Tells whether this stream respondsTo: reading protocol (next, next:into:startingAt:).
	Default is false, subclass should redefine if true."
	^false! !

!Xtream methodsFor: 'testing' stamp: 'nice 12/1/2009 21:11'!
isWriteable
	"Tells whether this stream respondsTo: writing protocol (nextPut:, next:putAll:startingAt:).
	Default is false, subclass should redefine if true."
	^false! !

!Xtream methodsFor: 'writing' stamp: 'nice 12/1/2009 21:23'!
next: count putAll: aCollection startingAt: startIndex
	self isWriteable
		ifTrue: [self subclassResponsibility]
		ifFalse: [self errorNotWriteable]! !

!Xtream methodsFor: 'writing' stamp: 'nice 12/1/2009 21:23'!
nextPut: anObject
	self isWriteable
		ifTrue: [self subclassResponsibility]
		ifFalse: [self errorNotWriteable]! !

!Xtream methodsFor: 'writing' stamp: 'nice 11/28/2009 17:49'!
nextPutAll: aCollection
	self next: aCollection size putAll: aCollection startingAt: 1! !

!Xtream methodsFor: 'positionning' stamp: 'nice 12/1/2009 21:21'!
position
	self isPositionable
		ifTrue: [self subclassResponsibility]
		ifFalse: [self errorNotPositionable]! !

!Xtream methodsFor: 'positionning' stamp: 'nice 12/1/2009 21:22'!
position: anInteger
	self isPositionable
		ifTrue: [self subclassResponsibility]
		ifFalse: [self errorNotPositionable]! !

!Xtream methodsFor: 'writing' stamp: 'nice 11/28/2009 18:35'!
print: anObject
	anObject printOn: self! !

!Xtream methodsFor: 'converting' stamp: 'nice 12/1/2009 21:33'!
readStream
	self isReadable ifTrue: [^self].
	^self contents readStream! !

!Xtream methodsFor: 'converting' stamp: 'nice 12/2/2009 12:36'!
readXtream
	"This message is for co-existence with Stream in an interim period."
	^self readStream! !

!Xtream methodsFor: 'positionning' stamp: 'nice 12/1/2009 21:22'!
skip: anInteger
	self isPositionable
		ifTrue: [self subclassResponsibility]
		ifFalse: [self errorNotPositionable]! !

!Xtream methodsFor: 'writing' stamp: 'nice 11/28/2009 18:35'!
store: anObject
	anObject storeOn: self! !

!Xtream methodsFor: 'converting' stamp: 'nice 12/1/2009 21:43'!
writeStream
	self isWriteable ifTrue: [^self].
	^self contents writeStream! !

!Xtream methodsFor: 'converting' stamp: 'nice 12/2/2009 12:36'!
writeXtream
	"This message is for co-existence with Stream in an interim period."
	^self writeStream! !

!StandardFileStream methodsFor: '*XTream-Core' stamp: 'nice 12/3/2009 23:00'!
readXtream
	^FileReadXtream new source: self! !

!StandardFileStream methodsFor: '*XTream-Core' stamp: 'nice 12/3/2009 23:02'!
writeXtream
	rwmode ifFalse: [self error: 'this file is readOnly'].
	^FileWriteXtream new destination: self! !

!SequenceableCollection methodsFor: '*XTream-Core' stamp: 'nice 12/2/2009 12:00'!
appendXtream
	^ CollectionWriteXtream append: self! !

!SequenceableCollection methodsFor: '*XTream-Core' stamp: 'nice 12/2/2009 12:03'!
readXtream
	^ CollectionReadXtream read: self! !

!SequenceableCollection methodsFor: '*XTream-Core' stamp: 'nice 12/2/2009 11:58'!
writeXtream
	^ CollectionWriteXtream write: self! !
